#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
from threading import Event, Thread
import time

from psbp.client.driver.pyserialrs485_driver import PySerialRS485Driver
from psbp.client.framer import Framer
from psbp.client.handler import Handler
from psbp.current_msg import MsgCurrentReq, PSBP_MSG_CURRENT_ACK
from psbp.frequency_msg import PSBP_MSG_FREQ_ACK, MsgFreqReq
from psbp.i_max_msg import MsgImaxWReq, PSBP_MSG_I_MAX_W_ACK
from psbp.mode_msg import PSBP_MSG_MODE_R_ACK, PSBP_MSG_MODE_W_ACK, MsgModeWReq
from psbp.phase_msg import PSBP_MSG_PHASE_ACK
from psbp.pll_offset_msg import PSBP_MSG_PLL_OFFSET_ACK
from psbp.power_msg import PSBP_MSG_POWER_W_ACK, PSBP_MSG_POWER_R_ACK, MsgPowerWReq, MsgPowerRReq
from psbp.status_msg import PSBP_MSG_STATUS, MsgStatus
from psbp.temp_msg import PSBP_MSG_TEMP_ACK, MsgTempReq
from psbp.utils import PsbpSlave, PsbpFlag, PsbpMode
from psbp.version_msg import PSBP_MSG_VERSION_ACK, MsgVersionReq
from psbp.voltage_msg import PSBP_MSG_VOLTAGE_ACK, MsgVoltageReq

# Debug logger
_logger = logging.getLogger(__name__)

# Power Limiter zur Strombegrenzung

PSBP_DEBUG = False

MAX_PWR = 100       # Maximaler Wert für Leistungsanforderung
TST_PWR = 21        # Leistungsanforderung bei Selbsttest
MIN_PWR = 0         # Minimaler Wert für Leistungsanforderung

ERROR_STATE_RECOVERY_TIME = 5   # 5 Sekunden Recovery Zeit bis zum Verlassen der Fehlerzustands


class PowerUnit(object):
    """!
    Klasse für die Leistungseinheit (LE).
    Übernimmt die Kommunikation mit der LE.
    Selbsttest der LE.
    Status der LE.
    """
    def __init__(self, config_manager, error_handler):
        """!
        Initialisiert die LE
        @param config_manager: Config Manager
        @param error_handler: Error Handler
        """
        _logger.debug("Power Unit init")
        self._name = "Power Unit"

        # Default Werte
        self._state = "INIT"
        self._selftest = 0      # -1 = Fehler beim Selbsttest, 0 = Selbsttest nicht ausgeführt, 1 = Selbsttest läuft, 2 = Selbsttest i.O.
        self._ps_temperature = 0
        self._ps_voltage = 0
        self._ps_frequency = 0
        self._ps_current_eff = 0
        self._ps_current_peak = 0
        self._ps_over_current = False
        self._ps_over_temperature = False
        self._ps_hc_defect = False
        self._ps_pwr_set_value = MIN_PWR
        self._ps_pwr_actual_value = MIN_PWR
        self._ps_flag_error = False
        self._ps_error_counter = 0
        self._ps_relay_on = False
        self._ps_msg_flag = 0
        self._ps_phase = 0
        self._ps_pll_offset = 0
        self._actuator_max_current = 0
        self._actuator_mode = PsbpMode.OFF
        self._actuator_enabled = False
        self._ps_response_received = False

        # Errorhandler
        self._error_handler = error_handler
        self._error_state_recovery_time = 0

        # Konfiguration aus Config-Datei laden
        self._config_manager = config_manager
        self._config = config_manager.get_configfile_parser()
        self._read_from_config()

        # RS485 Kommunikation aufbauen
        self._driver = PySerialRS485Driver()
        self._driver.__enter__()
        self._framer = Framer(read=self._driver.read, write=self._driver.write, verbose=True)
        self._actuator_link = Handler(source=self._framer)
        self._actuator_link.__enter__()
        self._ps_link = self._actuator_link

        # Callbacks für empfangene Botschaften einrichten
        self._setup_actuator_callbacks()
        # LE initialisieren
        self._init_actuator()

        # Überwachungs Thread einrichten und starten
        self._stopped = Event()
        self._stopped.clear()
        self._process_time = 0.2  # 200 msec process to calculate everything

        self._powerunit_thread = Thread(target=self._process, name=self._name)
        self._powerunit_thread.setDaemon(True)

        self._powerunit_thread.start()
        _logger.info("Power Unit thread started")

    def exit(self):
        """!
        Deaktiviert LE und beendet Thread
        """
        # Sofortiges Ausschalten
        self.disablePowerOutput()
        # Überwachungstask beenden
        self._stopped.set()
        self._framer.breakiter()
        time.sleep(self._process_time)
        _logger.debug("Power Unit exit()")

    def enablePowerOutput(self):
        """!
        Schaltet die LE ein und setzt den entsprechenden Modus (PWM / Phasenanschnitt).
        Initiale Leistungsanforderung ist 0
        @return True - Befehl wurde ausgeführt, False - Befehl wurde nicht ausgeführt
        """
        if self._state == "NORMAL":
            # set power to 0 %
            self._ps_link(MsgPowerWReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST, set_value=MIN_PWR))
            # set mode to 0
            self._ps_link(MsgModeWReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST, mode=self._actuator_mode))
            self._actuator_enabled = True
            return True
        else:
            return False

    def disablePowerOutput(self):
        """!
        Schaltet die LE aus.
        Setzt Leistungsanforderung auf 0
        @return True - Befehl wurde ausgeführt, False - Befehl wurde nicht ausgeführt
        """
        if self._state == "NORMAL":
            # set power to 0 %
            self._ps_link(MsgPowerWReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST, set_value=MIN_PWR))
            # set mode to 0
            self._ps_link(MsgModeWReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST, mode=PsbpMode.OFF))
            self._actuator_enabled = False
            return True
        else:
            return False

    def setPowerOutput(self, value):
        """!
        Setzt die Leistungsanforderung an die LE.
        @param value: Anzufordernde Leistung (0..100)
        @return True - Befehl wurde ausgeführt, False - Befehl wurde nicht ausgeführt
        """
        if self._state == "NORMAL":
            if PSBP_DEBUG and not (MIN_PWR <= value <= MAX_PWR):
                _logger.warning("requested value not in range: {}".format(value))

            lim_val = value
            lim_val = min(lim_val, MAX_PWR)                         # Gesamtwert nach oben begrenzen
            lim_val = max(lim_val, MIN_PWR)                         # Gesamtwert nach unten begrenzen

            self._ps_link(MsgPowerWReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST, set_value=lim_val))
            return True
        else:
            return False

    def startSelftest(self):
        """!
        Startet den Selbsttest der LE.
        @return True - Befehl wurde ausgeführt, False - Befehl wurde nicht ausgeführt
        """
        if self._state != "SELF_TEST_CHECK_POWER_OUTPUT":
            self._selftest = 1
            return True
        else:
            return False

    def getSelftestState(self):
        """!
        Gibt den Status des Selbsttests zurück.
        @return  -1 = Fehler beim Selbsttest, 0 = Selbsttest nicht ausgeführt, 1 = Selbsttest läuft, 2 = Selbsttest i.O.
        """
        return self._selftest

    # Statusinformationen
    @property
    def getPhase(self):
        """!
        Gibt die Phasensynchronisation der Power Stage zurück.
        @return Wert der Phasensynchronisation
        """
        return self._ps_phase

    @property
    def getFlagError(self):
        """!
        Gibt das Fehler Flag der Power Stage zurück.
        @return True - Fehler in LE liegt vor, False - kein Fehler
        """
        return self._ps_flag_error

    @property
    def getFlagRelayOn(self):
        """!
        Gibt den Status des Leistungsrelais zurück.
        @return False - Relais offen, True - Relais geschlossen / aktiv
        """
        return self._ps_relay_on

    @property
    def getTemperature(self):
        """!
        Gibt die Temperatur der Leistungseinheit zurück.
        @return Temperatur der Leistungseinheit [K]
        """
        return self._ps_temperature

    @property
    def getOverTemperature(self):
        """!
        Status der Übertemperaturerkennung.
        @return False - Temperatur normal, True - Temperatur zu hoch
        """
        return self._ps_over_temperature

    @property
    def getCurrentEff(self):
        """!
        Gibt den Effektivwert des Stroms zurück.
        @return Effektivwert des Stroms [A]
        """
        return self._ps_current_eff

    @property
    def getOverCurrent(self):
        """!
        Status der Überstromerkennung.
        @return False - kein Überstrom, True - Überstrom
        """
        return self._ps_over_current

    @property
    def getHCDefect(self):
        """!
        Status Heizkreis defekt.
        @return False - kein Defekt, True - Defekt
        """
        return self._ps_hc_defect

    @property
    def getCurrentPeak(self):
        """!
        Gibt den Wert des Spitzenstroms zurück.
        @return Spitzenstrom [A]
        """
        return self._ps_current_peak

    @property
    def getPwrSetValue(self):
        """!
        Gibt den Wert der angeforderten Leistung zurück.
        @return Angeforderten Leistung (0..100)
        """
        return self._ps_pwr_set_value

    @property
    def getPwrActualValue(self):
        """!
        Gibt den Wert der aktuell gestellten Leistung zurück.
        @return Aktuell gestellte Leistung (0..100)
        """
        return self._ps_pwr_actual_value

    @property
    def getMainsFrequency(self):
        """!
        Gibt den Wert der Netzfrequenz zurück.
        @return Netzfrequenz in [Hz]
        """
        return self._ps_frequency

    @property
    def getMainsVoltage(self):
        """!
        Gibt den Wert der Netzspannung zurück.
        @return Netzspannung in [V]
        """
        return self._ps_voltage

    def _read_from_config(self):
        """!
        Liest die Konfiguration des Ofens aus dem Config File.
        """
        # Betriebsmodus Leistungsteil
        mode = self._config_manager.get_config_file_value('power_unit', 'operation-mode')
        if mode == 'pulse':
            self._actuator_mode = PsbpMode.PULSE
        elif mode == 'phase':
            self._actuator_mode = PsbpMode.PHASE
        else:
            self._actuator_mode = PsbpMode.OFF

        # Strombegrenzung Leistungsteil
        self._actuator_max_current = int(self._config_manager.get_config_file_value('power_unit', 'peak-current-max'))

    def _init_actuator(self):
        """!
        Initialisiert die LE.
        - Statusabfrage
        - Versionsabfrage
        - Spitzenstrombegrenzung
        - Temperaturabfrage
        - Leistung auf 0
        - Relais aus
        """
        # set power to 0 %
        self._ps_link(MsgPowerWReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST, set_value=MIN_PWR))
        # set mode to 0
        self._ps_link(MsgModeWReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST, mode=PsbpMode.OFF))
        # send status request
        self._ps_link(MsgStatus(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST))
        # send version request
        self._ps_link(MsgVersionReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST))
        # send imax W request
        self._ps_link(MsgImaxWReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST, set_value=self._actuator_max_current))
        # send temp request
        self._ps_link(MsgTempReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST))
        self._ps_response_received = False

    # RS485 Callbacks
    def _setup_actuator_callbacks(self):
        """!
        Definiert die Callbacks für die RS485 Verbindung zum Leistungsteil.
        """
        self._ps_link.add_callback(self._version_callback, PSBP_MSG_VERSION_ACK)
        self._ps_link.add_callback(self._mode_r_callback, PSBP_MSG_MODE_R_ACK)
        self._ps_link.add_callback(self._mode_w_callback, PSBP_MSG_MODE_W_ACK)
        self._ps_link.add_callback(self._power_r_callback, PSBP_MSG_POWER_R_ACK)
        self._ps_link.add_callback(self._power_w_callback, PSBP_MSG_POWER_W_ACK)
        self._ps_link.add_callback(self._status_callback, PSBP_MSG_STATUS)
        self._ps_link.add_callback(self._temp_callback, PSBP_MSG_TEMP_ACK)
        self._ps_link.add_callback(self._current_callback, PSBP_MSG_CURRENT_ACK)
        self._ps_link.add_callback(self._voltage_callback, PSBP_MSG_VOLTAGE_ACK)
        self._ps_link.add_callback(self._frequency_callback, PSBP_MSG_FREQ_ACK)
        self._ps_link.add_callback(self._i_max_w_callback, PSBP_MSG_I_MAX_W_ACK)
        self._ps_link.add_callback(self._phase_callback, PSBP_MSG_PHASE_ACK)
        self._ps_link.add_callback(self._pll_offset_callback, PSBP_MSG_PLL_OFFSET_ACK)

    # Callbacks für empfangene Botschaften
    def _status_callback(self, msg, **metadata):
        """!
        Status Callback.
        """
        if msg.source_address is PsbpSlave.ONE:
            self._ps_response_received = True
            if PSBP_DEBUG:
                _logger.debug("status ack received: {}".format(msg.flag))

            self._ps_msg_flag = msg.flag

            self._ps_relay_on = True if self._ps_msg_flag & PsbpFlag.RELAY_ON == PsbpFlag.RELAY_ON else False

            self._ps_flag_error = False
            self._ps_hc_defect = False
            self._ps_over_current = False
            self._ps_over_temperature = False

            if self._ps_msg_flag & PsbpFlag.ERROR == PsbpFlag.ERROR:
                if self._ps_error_counter < 3:
                    self._ps_error_counter += 1     # Fehlerzähler erhöhen
                else:
                    self._ps_flag_error = True
                    _logger.debug("error flag active")
                    # Recovery Timer setzen
                    self._error_state_recovery_time = time.time() + ERROR_STATE_RECOVERY_TIME
            else:
                self._ps_error_counter = 0

            if self._ps_msg_flag & PsbpFlag.HC_DEFECT == PsbpFlag.HC_DEFECT:
                self._ps_hc_defect = True
                _logger.debug("heater defect flag active")
                # Recovery Timer setzen
                self._error_state_recovery_time = time.time() + ERROR_STATE_RECOVERY_TIME

            if self._ps_msg_flag & PsbpFlag.OVER_CURRENT == PsbpFlag.OVER_CURRENT:
                self._ps_over_current = True
                _logger.debug("over current flag active")
                # Recovery Timer setzen
                self._error_state_recovery_time = time.time() + ERROR_STATE_RECOVERY_TIME

            if self._ps_msg_flag & PsbpFlag.OVER_TEMPERATURE == PsbpFlag.OVER_TEMPERATURE:
                self._ps_over_temperature = True
                _logger.debug("over temperature flag active")
                # Recovery Timer setzen
                self._error_state_recovery_time = time.time() + ERROR_STATE_RECOVERY_TIME

    @staticmethod
    def _i_max_w_callback(msg, **metadata):
        """!
        Spitzenstrombegrenzung Callback.
        """
        if msg.source_address is PsbpSlave.ONE:
            if PSBP_DEBUG:
                _logger.debug("imax w ack")
                _logger.debug("actual_value:{}".format(msg.actual_value))

    def _current_callback(self, msg, **metadata):
        """!
        Strom Callback.
        """
        if msg.source_address is PsbpSlave.ONE:
            if PSBP_DEBUG:
                _logger.debug("current ack")
                _logger.debug("current_effective:{}".format(msg.current_effective))
                _logger.debug("current_peak:{}".format(msg.current_peak))

            self._ps_current_eff = msg.current_effective / 1000.0
            self._ps_current_peak = msg.current_peak / 1000.0

    def _version_callback(self, msg, **metadata):
        """!
        Versions Callback.
        """
        if msg.source_address is PsbpSlave.ONE:
            if PSBP_DEBUG:
                _logger.debug("version ack")
                _logger.debug("hw version:{}".format(msg.hw_version))
                _logger.debug("sw version:{}".format(msg.sw_version))

            self._config_manager.set_production_file_value("powerunit", "gs1-sw-power-section-1",
                                                           "{}".format(msg.sw_version))
            self._config_manager.set_production_file_value("powerunit", "gs1-hw-power-section-1",
                                                           "{}".format(msg.hw_version))

    def _temp_callback(self, msg, **metadata):
        """!
        Temperatur Callback.
        """
        if msg.source_address is PsbpSlave.ONE:
            if PSBP_DEBUG:
                _logger.debug("temp ack")
                _logger.debug("ps temp:{}".format(msg.temp))

            self._ps_temperature = msg.temp / 10.0

    def _power_r_callback(self, msg, **metadata):
        """!
        Leistungs-Abfrage Callback.
        """
        if msg.source_address is PsbpSlave.ONE:
            self._ps_pwr_set_value = msg.set_value
            self._ps_pwr_actual_value = msg.actual_value
            if PSBP_DEBUG:
                _logger.debug("power r ack")
                _logger.debug("set_value:{}".format(self._ps_pwr_set_value))
                _logger.debug("actual_value:{}".format(self._ps_pwr_actual_value))

    def _power_w_callback(self, msg, **metadata):
        """!
        Leistungs-Anforderung Callback.
        """
        if msg.source_address is PsbpSlave.ONE:
            self._ps_pwr_set_value = msg.set_value
            self._ps_pwr_actual_value = msg.actual_value
            if PSBP_DEBUG:
                _logger.debug("power w ack")
                _logger.debug("set_value:{}".format(self._ps_pwr_set_value))
                _logger.debug("actual_value:{}".format(self._ps_pwr_actual_value))

    @staticmethod
    def _mode_r_callback(msg, **metadata):
        """!
        Modus-Abfrage Callback.
        """
        if msg.source_address is PsbpSlave.ONE:
            if PSBP_DEBUG:
                _logger.debug("mode r ack")
                _logger.debug("mode:{}".format(msg.mode))

    @staticmethod
    def _mode_w_callback(msg, **metadata):
        """!
        Modus-Setzen Callback.
        """
        if msg.source_address is PsbpSlave.ONE:
            if PSBP_DEBUG:
                _logger.debug("mode w ack")
                _logger.debug("mode:{}".format(msg.mode))

    def _voltage_callback(self, msg, **metadata):
        """!
        Netzspannung Callback.
        """
        if msg.source_address is PsbpSlave.ONE:
            if PSBP_DEBUG:
                _logger.debug("voltage ack")
                _logger.debug("voltage:{}".format(msg.voltage))
            self._ps_voltage = msg.voltage / 100.0  # voltage:23263

    def _frequency_callback(self, msg, **metadata):
        """!
        Netzfrequenz Callback.
        """
        if msg.source_address is PsbpSlave.ONE:
            if PSBP_DEBUG:
                _logger.debug("frequency ack")
                _logger.debug("frequency:{}".format(msg.frequency))
            self._ps_frequency = msg.frequency / 10.0

    def _phase_callback(self, msg, **metadata):
        """!
        Phasen Synchronisierung Callback.
        """
        if msg.source_address is PsbpSlave.ONE:
            if PSBP_DEBUG:
                _logger.debug("phase ack")
                _logger.debug("phase:{}".format(msg.phase))
            self._ps_phase = msg.phase

    def _pll_offset_callback(self, msg, **metadata):
        """!
        PLL Callback.
        """
        if msg.source_address is PsbpSlave.ONE:
            if PSBP_DEBUG:
                _logger.debug("pll offset ack")
                _logger.debug("pll_offset:{}".format(msg.pll_offset))
            self._ps_pll_offset = msg.pll_offset

    def _process(self):
        """!
        Überwachungs Task der Leistungseinheit. Wird alle 200ms ausgeführt.
        """
        rate_time_logging = 1
        relais_state_mismatch = 0
        self_test_start_time = time.time()
        error_entry = True

        _logger.debug("{} process started".format(self._name))

        while not self._stopped.wait(self._process_time):
            process_start_time = time.time()

            ##########################################################
            # Statemachine                                           #
            ##########################################################
            if self._state == "INIT":
                if self._selftest == 1:
                    _logger.info("self test started")
                    self._error_handler.reset_error("F30")
                    self._error_handler.reset_error("F32")
                    # Zeit nehmen
                    self_test_start_time = time.time()
                    # LE Modus setzen, für Selbsttest LE auf 10
                    self._ps_link(MsgModeWReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST, mode=self._actuator_mode))
                    self._ps_link(MsgPowerWReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST, set_value=TST_PWR))
                    # Zustand wechseln
                    self._state = "SELF_TEST_CHECK_POWER_OUTPUT"

            elif self._state == "SELF_TEST_CHECK_POWER_OUTPUT":
                if (self_test_start_time + 5) <= time.time():     # LE soll 5s lang aktiv sein
                    # Stromfluss ist n.i.O. oder
                    # eines der Fehlerflags wurde gesetzt
                    if self._ps_hc_defect:
                        self._selftest = -1
                        self._error_handler.set_error("F30")
                        _logger.error("self test failed, goto ERROR")
                        self._state = "ERROR"

                    if (self._ps_flag_error or
                            self._ps_over_current or
                            self._ps_over_temperature):
                        self._selftest = -1
                        self._error_handler.set_error("F32")
                        _logger.error("self test failed, goto ERROR")
                        self._state = "ERROR"

                    if not self._selftest == -1:
                        self._selftest = 2
                        _logger.info("self test passed, goto NORMAL")
                        self._state = "NORMAL"
                        relais_state_mismatch = 0

                    # Leistung auf 0, Relais aus
                    self._ps_link(MsgPowerWReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST, set_value=MIN_PWR))
                    self._ps_link(MsgModeWReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST, mode=PsbpMode.OFF))

            elif self._state == "NORMAL":
                '''
                Ein Stromausfall wird an dieser Stelle nicht sicher erkannt. Wenn der Strom ausfällt, kommt hier kein Fehler an.
                Nachdem der Strom wieder da ist, geht es der LE "gut". Trotzdem muss Sie initialisiert werden.
                Es muss ein Mechanismus her, den den Stromausfall erkennt und danach die LE initialisiert.
                Der Mechanismus darf sich nicht darauf verlassen, dass die LE vor oder nach dem Stromausfall vernünftige Werte sendet.
                Vielleicht ist es möglich den Relais Status gegen den erwarteten Status zu überprüfen.
                  - Soll das Relais ein sein,
                  - und es ist nicht aktiv
                  - und es sind keine Fehler vorhanden, muss die LE erneut initialisiert werden.
                '''
                if self._ps_hc_defect:
                    self._error_handler.set_error("F30")
                    _logger.error("error flag is set, goto ERROR")
                    # In ERROR State wechseln
                    self._state = "ERROR"
                elif (self._ps_flag_error or
                      self._ps_over_current or
                      self._ps_over_temperature):
                        self._error_handler.set_error("F32")
                        _logger.error("error flag is set, goto ERROR")
                        # In ERROR State wechseln
                        self._state = "ERROR"
                elif (self._actuator_enabled and
                      not self._ps_relay_on):
                        relais_state_mismatch += 1
                        _logger.debug("actuator is expected to be active, but relais is off")
                        if relais_state_mismatch > 5:
                            _logger.error("relais state mismatch, goto ERROR")
                            # In ERROR State wechseln
                            self._state = "ERROR"
                elif self._selftest == 1:
                    _logger.info("self test started in normal state")
                    # Zustand wechseln
                    self._state = "INIT"
                else:
                    relais_state_mismatch = 0

            elif self._state == "ERROR":
                if error_entry == True:
                    error_entry = False
                    # Recovery Timer setzen
                    self._error_state_recovery_time = time.time() + ERROR_STATE_RECOVERY_TIME
                    # Leistung auf 0, Relais aus
                    self._ps_link(
                        MsgPowerWReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST, set_value=MIN_PWR))
                    self._ps_link(
                        MsgModeWReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST, mode=PsbpMode.OFF))

                if self._selftest == 1:
                    # LE initialisieren
                    self._init_actuator()
                    _logger.info("self test started from error state")
                    error_entry = True
                    # Zustand wechseln
                    self._state = "INIT"
                elif self._selftest == 2:
                    if (self._ps_hc_defect or
                      self._ps_flag_error or
                      self._ps_over_current or
                      self._ps_over_temperature):
                        # Recovery Timer neu setzen
                        self._error_state_recovery_time = time.time() + ERROR_STATE_RECOVERY_TIME
                        _logger.info("error flags still active, restart recovery timer")
                    elif self._error_state_recovery_time < time.time():
                        # LE initialisieren
                        self._init_actuator()
                        if self._actuator_enabled:
                            self._ps_link(MsgModeWReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST, mode=self._actuator_mode))
                        _logger.info("recovery time expired, goto NORMAL")
                        self._state = "NORMAL"
                        error_entry = True
                        relais_state_mismatch = 0

            ##########################################################
            # Statusabfragen                                         #
            ##########################################################

            # if not self._ps_response_received:
            #     _logger.error("no status response received from power stage")
            # send status request
            self._ps_link(MsgStatus(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST))
            self._ps_response_received = False

            if rate_time_logging == 1:
                # read power section temperature
                self._ps_link(MsgTempReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST))
                # next
                rate_time_logging = 2

            elif rate_time_logging == 2:
                # send voltage request
                self._ps_link(MsgVoltageReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST))
                # send frequency request
                self._ps_link(MsgFreqReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST))
                # next
                rate_time_logging = 3

            elif rate_time_logging == 3:
                # send current request
                self._ps_link(MsgCurrentReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST))
                # send current power value request
                self._ps_link(MsgPowerRReq(receiver_address=PsbpSlave.ONE, flag=PsbpFlag.ACK_REQUEST))
                # next
                rate_time_logging = 1

            delta_process_time = time.time() - process_start_time
            wait_time = self._process_time - delta_process_time
            if wait_time <= 0:
                if __debug__:
                    _logger.warning("no wait time:{}".format(delta_process_time))

        _logger.debug("{} process stopped".format(self._name))
